from dataclasses import dataclass

import numpy as np
from astropy.coordinates import Longitude, SkyCoord
from astropy.wcs import WCS


@dataclass
class Sources:
    """A collection of sources with their sky coordinates and fluxes.

    Examples
    --------

    Create a Sources instance from arrays:

    >>> from cabaret.sources import Sources
    >>> import numpy as np
    >>> from astropy.coordinates import SkyCoord
    >>> coords = np.array([[10.64, 41.26], [10.68, 41.22]])
    >>> fluxes = np.array([169435.6, 52203.9])
    >>> sources = Sources(SkyCoord(coords, unit='deg'), fluxes)
    >>> sources
    Sources(coords=<SkyCoord (ICRS): (ra, dec) in deg
        [(10.64, 41.26), (10.68, 41.22)]>, fluxes=array([169435.6,  52203.9]))

    """

    coords: SkyCoord
    """SkyCoords instance with the RA and DEC coordinates of the sources."""
    fluxes: np.ndarray
    """An array of shape (n,) containing the fluxes of the sources."""

    def __post_init__(self):
        if not isinstance(self.coords, SkyCoord):
            raise ValueError("coords must be an instance of SkyCoord.")
        if not isinstance(self.fluxes, np.ndarray):
            try:
                self.fluxes = np.array(self.fluxes)
            except Exception:
                raise ValueError("fluxes must be an instance of np.ndarray.")
        if self.coords.size != self.fluxes.size:
            raise ValueError("coords and fluxes must have the same length.")

    @property
    def ra(self) -> Longitude:
        """Right Ascension coordinates of the sources."""
        return self.coords.ra  # type: ignore

    @property
    def dec(self) -> Longitude:
        """Declination coordinates of the sources."""
        return self.coords.dec  # type: ignore

    @property
    def center(self) -> tuple[float, float]:
        """Midpoint RA and DEC of the sources in degrees."""
        ra_min, ra_max = self.ra.deg.min(), self.ra.deg.max()  # type: ignore
        ra_range = ra_max - ra_min

        if ra_range > 180:
            ra_shifted = (self.ra.deg + 180) % 360  # type: ignore
            ra_min, ra_max = ra_shifted.min(), ra_shifted.max()  # type: ignore
            ra_center = (ra_min + ra_max) / 2 - 180
            ra_center %= 360
        else:
            ra_center = (ra_min + ra_max) / 2

        dec_center = (self.dec.deg.min() + self.dec.deg.max()) / 2  # type: ignore
        return ra_center, dec_center

    def to_pixel(self, wcs: WCS) -> np.ndarray:
        """Convert the RA-DEC coordinates to pixel coordinates using the given WCS.

        Parameters
        ----------
        wcs : astropy.wcs.WCS
            The WCS object used for the conversion.

        Returns
        -------
        np.ndarray
            An array of shape (n, 2) containing the pixel coordinates of the sources.
        """
        return np.array(self.coords.to_pixel(wcs))

    def __len__(self) -> int:
        return len(self.fluxes)

    def __getitem__(self, key) -> "Sources":
        new_fluxes = np.asarray(self.fluxes)[key]
        if np.isscalar(new_fluxes):
            new_fluxes = np.array([new_fluxes])

        new_coords = self.coords[key]

        return Sources(new_coords, new_fluxes)  # type: ignore

    @classmethod
    def from_arrays(
        cls,
        ra: np.ndarray | list,
        dec: np.ndarray | list,
        fluxes: np.ndarray | list,
        units: str = "deg",
    ) -> "Sources":
        """Create a Sources instance from separate RA and DEC arrays.

        Parameters
        ----------
        ra : np.ndarray
            An array of shape (n,) containing the RA coordinates of the sources in deg.
        dec : np.ndarray
            An array of shape (n,) containing the DEC coordinates of the sources in deg.
        fluxes : np.ndarray
            An array of shape (n,) containing the fluxes of the sources.
        **kwargs
            Additional keyword arguments passed to the Sources constructor.

        Returns
        -------
        Sources
            A Sources instance.
        """
        if not isinstance(ra, np.ndarray):
            try:
                ra = np.array(ra)
            except Exception:
                raise ValueError("ra must be an instance of np.ndarray.")
        if not isinstance(dec, np.ndarray):
            try:
                dec = np.array(dec)
            except Exception:
                raise ValueError("dec must be an instance of np.ndarray.")
        if ra.shape != dec.shape:
            raise ValueError("ra and dec must have the same shape.")

        parameters = {
            "coords": SkyCoord(ra=ra, dec=dec, unit=units),
            "fluxes": fluxes,
        }
        return cls(**(parameters))

    @classmethod
    def get_test_sources(cls) -> "Sources":
        """Return a simple test Sources instance."""
        coords = SkyCoord(
            ra=[12.29611593, 12.29929654, 12.33757534, 12.34247842, 12.29354464],
            dec=[30.45675318, 30.44855405, 30.42613357, 30.48059276, 30.47310728],
            unit="deg",
        )
        fluxes = np.array([307220.0, 64271.0, 61002.0, 43466.0, 9239.0])
        return cls(coords, fluxes=fluxes)



import pytest

from cabaret import Camera, generate_image

from .utils import has_internet


@pytest.mark.skipif(not has_internet(), reason="Requires internet")
def test_simple():
    camera = Camera(width=100, height=100)
    ra, dec = 323.36152, -0.82325
    _ = generate_image(ra, dec, 1, camera=camera)


@pytest.mark.skipif(not has_internet(), reason="Requires internet")
def test_timeout():
    camera = Camera(width=100, height=100)
    ra, dec = 323.36152, -0.82325

    with pytest.raises(TimeoutError):
        _ = generate_image(ra, dec, 1, camera=camera, timeout=0.1)



import copy
from dataclasses import asdict, dataclass
from datetime import datetime
from pathlib import Path
from typing import Any

import numpy.random
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.wcs import WCS

from cabaret.camera import Camera
from cabaret.fits_manager import FITSManager
from cabaret.focuser import Focuser
from cabaret.image import generate_image, generate_image_stack
from cabaret.queries import Filters
from cabaret.site import Site
from cabaret.sources import Sources
from cabaret.telescope import Telescope


@dataclass
class Observatory:
    """
    Observatory configuration.

    Examples
    --------
    >>> from datetime import datetime, UTC
    >>> dateobs = datetime.now(UTC)
    >>> from cabaret.observatory import Observatory
    >>> observatory = Observatory()

    Query Gaia for sources and generate an image:

    >>> image = observatory.generate_image(
    ...     ra=12.3323, dec=30.4343, exp_time=10, dateobs=dateobs, seed=0
    ... )

    Or using a set of predefined sources:

    >>> from cabaret.sources import Sources
    >>> sources = Sources.from_arrays(
    ...     ra=[10.64, 10.68], dec=[10.68, 41.22], fluxes=[169435.6, 52203.9]
    ... )
    >>> img = observatory.generate_image(
    ...     ra=sources.ra.deg.mean(),
    ...     dec=sources.dec.deg.mean(),
    ...     exp_time=10,
    ...     seed=0,
    ...     sources=sources,
    ... )

    If you have matplotlib installed, you can visualize the image using cabaret's plot
    utility:

    >>> import matplotlib.pyplot as plt
    >>> from cabaret.plot import plot_image
    >>> _ = plot_image(image, title="Simulated Image")
    >>> plt.show()
    """

    name: str
    """Observatory name."""

    camera: Camera
    """Camera configuration."""

    focuser: Focuser
    """Focuser configuration."""

    telescope: Telescope
    """Telescope configuration."""

    site: Site
    """Site configuration."""

    def __init__(
        self,
        name: str = "Observatory",
        camera: Camera | dict = Camera(),
        focuser: Focuser | dict = Focuser(),
        telescope: Telescope | dict = Telescope(),
        site: Site | dict = Site(),
    ):
        if isinstance(camera, dict):
            camera = Camera(**camera)
        if isinstance(focuser, dict):
            focuser = Focuser(**focuser)
        if isinstance(telescope, dict):
            telescope = Telescope(**telescope)
        if isinstance(site, dict):
            site = Site(**site)

        self.name = name
        self.camera = camera
        self.focuser = focuser
        self.telescope = telescope
        self.site = site
        self.__post_init__()

    def __post_init__(self):
        if not isinstance(self.camera, Camera):
            raise ValueError("camera must be an instance of Camera.")
        if not isinstance(self.focuser, Focuser):
            raise ValueError("focuser must be an instance of Focuser.")
        if not isinstance(self.telescope, Telescope):
            raise ValueError("telescope must be an instance of Telescope.")
        if not isinstance(self.site, Site):
            raise ValueError("site must be an instance of Site.")

    def generate_image(
        self,
        ra: float,
        dec: float,
        exp_time: float,
        dateobs: datetime | None = None,
        light: int = 1,
        filter_band: Filters | str = Filters.G,
        airmass: float = 1.5,
        n_star_limit: int = 2000,
        rng: numpy.random.Generator = numpy.random.default_rng(),
        seed: int | None = None,
        timeout: float | None = None,
        sources: Sources | None = None,
        wcs: WCS | None = None,
        fwhm_multiplier: float = 5.0,
    ) -> numpy.ndarray:
        """Generate a simulated image of the sky.

        Parameters
        ----------
        ra : float
            Right ascension of the center of the image in degrees.
        dec : float
            Declination of the center of the image in degrees.
        exp_time : float
            Exposure time in seconds.
        dateobs : datetime, optional
            Observation date and time in UTC.
        light : int, optional

        filter_band : Filters or str, optional
            Photometric filter to use for the simulation (default: Filters.G).
        airmass : float, optional
            Airmass value for the observation (default: 1.5).
        n_star_limit : int, optional
            Maximum number of stars to include in the image.
        rng : numpy.random.Generator, optional
            Random number generator.
        seed : int, optional
            Random number generator seed.
        timeout : float, optional
            The maximum time to wait for the Gaia query to complete, in seconds.
            If None, there is no timeout. By default, it is set to None.
        sources : Sources, optional
            A collection of sources with their sky coordinates and fluxes.
            If provided, these sources will be used instead of querying Gaia.
        wcs : WCS or None, optional
            World Coordinate System information for the image.
        fwhm_multiplier : float, optional
            Multiplier to determine the rendering radius around each star
            (default: 5.0).
        """
        return generate_image(
            ra=ra,
            dec=dec,
            exp_time=exp_time,
            dateobs=dateobs,
            light=light,
            camera=self.camera,
            focuser=self.focuser,
            telescope=self.telescope,
            site=self.site,
            filter_band=filter_band,
            airmass=airmass,
            n_star_limit=n_star_limit,
            rng=rng,
            seed=seed,
            timeout=timeout,
            sources=sources,
            wcs=wcs,
            fwhm_multiplier=fwhm_multiplier,
        )

    def generate_image_stack(
        self,
        ra: float,
        dec: float,
        exp_time: float,
        dateobs: datetime | None = None,
        light: int = 1,
        filter_band: Filters | str = Filters.G,
        airmass: float = 1.5,
        n_star_limit: int = 2000,
        rng: numpy.random.Generator = numpy.random.default_rng(),
        seed: int | None = None,
        timeout: float | None = None,
        sources: Sources | None = None,
        convert_all_to_adu: bool = True,
        wcs: WCS | None = None,
        fwhm_multiplier: float = 5.0,
    ) -> numpy.ndarray:
        """
        Generate a stack of images from different stages in the image simulation
        pipeline.

        From first to last, the images are:
        1. Base image with bias, dark, and flat applied.
        2. Astronomical image with sources, sky background, and noise.
        3. Final image with pixel defects applied.

        Parameters
        ----------
        ra : float
            Right ascension of the image center (degrees).
        dec : float
            Declination of the image center (degrees).
        exp_time : float
            Exposure time in seconds.
        dateobs : datetime, optional
            Observation date and time (default: now, UTC).
        light : int, optional
            If 1, simulate light exposure; if 0, simulate dark exposure.
        camera : Camera, optional
            Camera configuration.
        focuser : Focuser, optional
            Focuser configuration.
        telescope : Telescope, optional
            Telescope configuration.
        site : Site, optional
            Observatory site configuration.
        filter_band : Filters or str, optional
            The filter to use for the flux column. Default is "G".
        airmass : float, optional
            Airmass value for the observation (default: 1.5).
        n_star_limit : int, optional
            Maximum number of stars to simulate.
        rng : numpy.random.Generator, optional
            Random number generator.
        seed : int or None, optional
            Seed for the random number generator.
        timeout : float or None, optional
            Timeout for Gaia query.
        sources : Sources or None, optional
            Precomputed sources to use instead of querying Gaia.
        convert_all_to_adu : bool, optional
            Whether to convert all images to ADU. Default is False.
        wcs : WCS or None, optional
            World Coordinate System information for the image.
        fwhm_multiplier : float, optional
            Multiplier to determine the rendering radius around each star
            (default: 5.0).

        Returns
        -------
        np.ndarray
            Simulated image stack as a 3D array (uint16, shape (3, height, width)).
            The first slice is the base image, the second is the astronomical image,
            and the third is the ADU image with pixel defects applied.


        """
        return generate_image_stack(
            ra=ra,
            dec=dec,
            exp_time=exp_time,
            dateobs=dateobs,
            light=light,
            camera=self.camera,
            focuser=self.focuser,
            telescope=self.telescope,
            site=self.site,
            filter_band=filter_band,
            airmass=airmass,
            n_star_limit=n_star_limit,
            rng=rng,
            seed=seed,
            timeout=timeout,
            sources=sources,
            convert_all_to_adu=convert_all_to_adu,
            wcs=wcs,
            fwhm_multiplier=fwhm_multiplier,
        )

    def generate_fits_image(
        self,
        ra: float,
        dec: float,
        exp_time: float,
        file_path: str | Path | None = None,
        dateobs: datetime | None = None,
        light: int = 1,
        filter_band: Filters | str = Filters.G,
        airmass: float = 1.5,
        n_star_limit: int = 2000,
        rng: numpy.random.Generator = numpy.random.default_rng(),
        seed: int | None = None,
        timeout: float | None = None,
        sources: Sources | None = None,
        wcs: WCS | None = None,
        fwhm_multiplier: float = 5.0,
        user_header: dict[str, Any] | fits.Header | None = None,
        overwrite: bool = True,
    ) -> fits.HDUList:
        """Generate a simulated FITS image of the sky.

        Parameters
        ----------
        ra : float
            Right ascension of the center of the image in degrees.
        dec : float
            Declination of the center of the image in degrees.
        exp_time : float
            Exposure time in seconds.
        file_path : str or Path, optional
            If provided, the path to save the FITS file.
        user_header : dict or fits.Header, optional
            Additional header keywords to add.
        dateobs : datetime, optional
            Observation date and time in UTC.
        light : int, optional
            If 1, simulate light exposure; if 0, simulate dark exposure.
        filter_band : Filters or str, optional
            Photometric filter to use for the simulation (default: Filters.G).
        airmass : float, optional
            Airmass value for the observation (default: 1.5).
        n_star_limit : int, optional
            Maximum number of stars to include in the image.
        rng : numpy.random.Generator, optional
            Random number generator.
        seed : int, optional
            Random number generator seed.
        timeout : float, optional
            The maximum time to wait for the Gaia query to complete, in seconds.
            If None, there is no timeout. By default, it is set to None.
        sources : Sources, optional
            A collection of sources with their sky coordinates and fluxes.
            If provided, these sources will be used instead of querying Gaia.
        wcs : WCS or None, optional
            World Coordinate System information for the image.
        fwhm_multiplier : float, optional
            Multiplier to determine the rendering radius around each star
            (default: 5.0).
        overwrite : bool, optional
            Whether to overwrite existing file (default: True).

        Returns
        -------
        fits.HDUList
            The generated FITS HDU list.
        """
        image = generate_image(
            ra=ra,
            dec=dec,
            exp_time=exp_time,
            dateobs=dateobs,
            light=light,
            camera=self.camera,
            focuser=self.focuser,
            telescope=self.telescope,
            site=self.site,
            filter_band=filter_band,
            airmass=airmass,
            n_star_limit=n_star_limit,
            rng=rng,
            seed=seed,
            timeout=timeout,
            sources=sources,
            wcs=wcs,
            fwhm_multiplier=fwhm_multiplier,
        )

        if wcs is None:
            wcs = self.camera.get_wcs(SkyCoord(ra=ra, dec=dec, unit="deg"))

        hdu_list = FITSManager.to_hdu_list(
            observatory=self,
            image=image,
            user_header=user_header,
            ra=ra,
            dec=dec,
            exp_time=exp_time,
            dateobs=dateobs,
            wcs=wcs,
        )
        if file_path is not None:
            hdu_list = FITSManager.save(
                observatory=self,
                file_path=file_path,
                hdu_list=hdu_list,
                overwrite=overwrite,
            )
        return hdu_list

    def to_dict(self) -> dict:
        """Convert the Observatory configuration to a dictionary."""
        return asdict(self)

    @classmethod
    def from_dict(cls, config) -> "Observatory":
        """Create an Observatory instance from a configuration dictionary."""
        return cls(
            name=config.get("name", "Observatory"),
            camera=Camera(**config["camera"]),
            focuser=Focuser(**config.get("focuser", {})),
            telescope=Telescope(**config["telescope"]),
            site=Site(**config["site"]),
        )

    @classmethod
    def load_from_yaml(cls, file_path: str | Path) -> "Observatory":
        """Load Observatory configuration from a YAML file."""
        try:
            import yaml

            with open(file_path) as f:
                config = yaml.safe_load(f)

            return cls.from_dict(config)

        except ImportError:
            raise ImportError(
                "Please install PyYAML to load Observatory configuration from YAML."
            )
        except FileNotFoundError:
            raise FileNotFoundError(f"File not found: {file_path}")
        except Exception as e:
            raise Exception(f"Error loading Observatory configuration: {e}")

    def save_to_yaml(self, file_path: str | Path):
        """Save Observatory configuration to a YAML file."""
        try:
            import yaml

            with open(file_path, "w") as f:
                yaml.dump(self.to_dict(), f)

        except ImportError:
            raise ImportError(
                "Please install PyYAML to save Observatory configuration to YAML."
            )
        except Exception as e:
            raise Exception(f"Error saving Observatory configuration: {e}")

    def copy(self) -> "Observatory":
        """Create a deep copy of the Observatory instance."""
        return copy.deepcopy(self)



import importlib.metadata
import os
import sys

# Add the project source directory to the path so that autodoc can find the modules
sys.path.insert(0, os.path.abspath("../../src"))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "cabaret"
copyright = "2025, Peter Pedersen, Lionel Garcia, David Degen"
author = "Peter Pihlmann, Lionel Garcia, David Degen"

version = importlib.metadata.version("cabaret")

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_nb",
    # "numpydoc",
    "sphinx_copybutton",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx_design",
    "IPython.sphinxext.ipython_console_highlighting",
]

nb_render_image_options = {"align": "center"}

# Add mappings for intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_css_files = ["style.css"]
html_short_title = "cabaret"
html_title = f"{html_short_title}"

html_context = {
    "default_mode": "light",
}

html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/ppp-one/cabaret",
    "repository_branch": "main",
    "use_repository_button": True,
    "use_fullscreen_button": False,
    "use_issues_button": True,
    "use_download_button": False,
    "show_navbar_depth": 1,
    "navbar_end": ["navbar-icon-links"],
    "collapse_navigation": True,
}


# Auto-generate API documentation
autodoc_member_order = "bysource"
autodoc_typehints = "signature"
autoclass_content = "class"
autodoc_default_options = {"no-value": True}
autodoc_preserve_defaults = False

myst_enable_extensions = ["dollarmath", "colon_fence"]

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_use_ivar = True
napoleon_attr_annotations = True
